/*
        @@@@@    @@@@    @@@@   @@  @@  @@   @@  @@@@   @@  @@
        @@  @@  @@  @@  @@  @@  @@ @@   @@@ @@@ @@  @@  @@@ @@
        @@@@@   @@@@@@  @@      @@@@    @@ @ @@ @@@@@@  @@ @@@
        @@      @@  @@  @@  @@  @@ @@   @@   @@ @@  @@  @@  @@
        @@      @@  @@   @@@@   @@  @@  @@   @@ @@  @@  @@  @@

        @@       @@@@    @@@@    @@@@   @@      @@@@@    @@@@    @@@@
        @@      @@  @@  @@      @@      @@      @@      @@      @@
        @@      @@  @@   @@@@    @@@@   @@      @@@@     @@@@    @@@@
        @@      @@  @@      @@      @@  @@      @@          @@      @@
        @@@@@@   @@@@    @@@@    @@@@   @@@@@@  @@@@@    @@@@    @@@@

        @@  @@  @@@@@@  @@@@@   @@@@@    @@@@
        @@  @@    @@    @@  @@  @@      @@  @@
        @@  @@    @@    @@  @@  @@@@    @@  @@
         @@@@     @@    @@  @@  @@      @@  @@
          @@    @@@@@@  @@@@@   @@@@@    @@@@

         @@@@    @@@@   @@@@@   @@@@@    @@@@
        @@  @@  @@  @@  @@  @@  @@      @@  @@
        @@      @@  @@  @@  @@  @@@@    @@
        @@  @@  @@  @@  @@  @@  @@      @@  @@
         @@@@    @@@@   @@@@@   @@@@@    @@@@

        PackMan Lossless Video DeCoder

        RLE -> HUFFMAN => WaveLet Transform => Frame Difference => YCbCr => RGB
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>

#include "huffman.h"
#include "rle.h"

//#define VIDEO /*  -:    DOS, Win9x, WinXP.     VESA VBE */

#define u8 unsigned char
#define s8 signed char

#define u16 unsigned short int
#define s16 signed short int

#define u32 unsigned long int
#define s32 signed long int

#define W 160 /*   */
#define H 128 /*   */

#define NW 3  /*  WaveLet  */

#ifdef VIDEO

#include <dos.h>
#include <dpmi.h>
#include <go32.h>
#include <sys/farptr.h>

#include "Video.inc"
#include "Timer.inc"

#endif

u8 nkf; //No Key Frame

u8 Compress[H*W];       // #2   

s8 Frame[H][W];         // ,   -  #1   
s8 Frame_Save[3][H][W]; // 

s8 CacheIN[W];          //  WaveLet 
s8 CacheOUT[W];

u8 RGB[H][W][3];        //  

u16 HUF_Len; //  
u16 RLE_Len;

void YCbCrDecode(void)
{
 register u8 r,g,b;
 register s8 y,cb,cr;

 for(u32 j=0;j<H;j++)for(u32 i=0;i<W;i++)
 {
  y =Frame_Save[2][j][i];
  cb=Frame_Save[1][j][i];
  cr=Frame_Save[0][j][i];

  g=(u8)(y-((cb+cr)>>2)); //Reversible Inverse Color Transform: ,  _!!!   int(-x) != floor(-x)
  r=(u8)(cr+g);
  b=(u8)(cb+g);

  r<<=2; //Convert 6:6:6 to 8:8:8 (  PC   -)
  g<<=2;
  b<<=2;

  RGB[j][i][2]=b; //Swap R<=>B (  PC   -)
  RGB[j][i][1]=g;
  RGB[j][i][0]=r;

#ifdef VIDEO

  _farpokeb(VideoRAM,(((j*W)+i)*3)+2,r); //    PC
  _farpokeb(VideoRAM,(((j*W)+i)*3)+1,g);
  _farpokeb(VideoRAM,(((j*W)+i)*3)  ,b);

#endif

 }
}

void LE_GALL_53_I(u32 n) //Le Gall 5/3 Inverse WaveLet Transform(IWT)
{
 register u32 i;

 //Unpack
 for(i=0;i<(n>>1);i++)
 {
  CacheOUT[ i<<1   ]=CacheIN[i       ];
  CacheOUT[(i<<1)|1]=CacheIN[i+(n>>1)];
 }

 //Undo update
 for(i=2;i<n;i+=2)CacheOUT[i]-=(CacheOUT[i-1]+CacheOUT[i+1])/4;
 CacheOUT[0]-=CacheOUT[1]/2;

 //Undo predict
 for(i=1;i<n-2;i+=2)CacheOUT[i]+=(CacheOUT[i-1]+CacheOUT[i+1])/2;
 CacheOUT[n-1]+=CacheOUT[n-2];
}

void IWT_X(u32 n)
{
 u32 x,y;
 for(y=0;y<H>>n;y++)
 {
  for(x=0;x<W>>n;x++)CacheIN[x]=Frame[y][x];
  LE_GALL_53_I(W>>n);
  for(x=0;x<W>>n;x++)Frame[y][x]=CacheOUT[x];
 }
}

void IWT_Y(u32 n)
{
 u32 x,y;
 for(x=0;x<W>>n;x++)
 {
  for(y=0;y<H>>n;y++)CacheIN[y]=Frame[y][x];
  LE_GALL_53_I(H>>n);
  for(y=0;y<H>>n;y++)Frame[y][x]=CacheOUT[y];
 }
}

u8 ReadFrame(FILE *f)       //  
{
 static u8 state=0;         // : 0 -  , 1 -  
 static u16 head_pred;      //    
 u16 head;

 if(!state)                 //  
 {
  state=1;
  fread(Compress,1,10,f);   //     1- 
  head=*(u16*)&Compress[8];
 }
 else head=head_pred;       //,     

 if(!head)return 0;         //  (  )

 u8 hr=head>>15;            // 15 => 0:HUF 1:RLE(HUF)

 nkf=((head>>14)&1)^1;      // 14 => 1:KeyFrame 0:No KeyFrame

 head&=0x3FFF;
 head++;                    // 0..13 => Compress Frame Size: 1..16384

 if(hr)                     //  2  ,      (       )
 {
  RLE_Len=head;
  fread((u8*)Frame,1,head+2,f);
  head_pred=*(u16*)(((u32)&Frame)+head);
 }
 else
 {
  RLE_Len=0;
  HUF_Len=head;
  fread(Compress,1,head+2,f);
  head_pred=*(u16*)&Compress[head];
 }

 return 1;
}

int main(int argc,char** argv)
{
 u32 frame=0; // 

 if(argc!=3)
 {
  printf("\nUsage: decoder InFile OutFile\n");
  return -1;
 }

 FILE *fin=fopen(argv[1],"rb");
 if(fin==NULL)return -1;

 FILE *fout=fopen(argv[2],"wb");
 if(fout==NULL)return -1;

 printf("\n");

#ifdef VIDEO

 Video_Init();

#endif

 memset(Frame_Save,0,sizeof(Frame_Save));
 
 Decode:

 if(kbhit())goto End;

 for(s32 c=2;c>=0;c--) //Y Cb Cr  - 
 {
  if(!ReadFrame(fin))goto End; //    -  

  if(RLE_Len)HUF_Len=DecompressRLE((u8*)Frame,RLE_Len,Compress);   //   RLE(HUF)
  if(DecompressHuffman(Compress,HUF_Len,(u8*)Frame)!=W*H)          //    
  {

#ifndef VIDEO

   printf("Incorrect Decode! Press any key...\n");
   getch();

#endif

   goto End; //  ( )
  }

  for(s32 n=NW-1;n>=0;n--) //2D Inverse WaveLet Transform
  {
   IWT_Y(n);
   IWT_X(n);
  }

  if(nkf)for(u32 i=0;i<W*H;i++)((s8*)&Frame_Save[c])[i]+=((s8*)&Frame)[i]; //KeyFrame ? Frame : Frame Difference
  else memcpy((void*)&Frame_Save[c],Frame,sizeof(Frame));
 }

 YCbCrDecode(); //YCbCr => RGB

#ifdef VIDEO

 SDL_Delay(83); //12 FPS

#endif

 fwrite(RGB,1,sizeof(RGB),fout); //  

#ifndef VIDEO

 printf("Frame: %u \n",frame); //   
 frame++;

#endif

 goto Decode; //   

 End:

 fclose(fout);
 fclose(fin);

#ifdef VIDEO

 Video_DeInit();

#else

 printf("\nDecoding finished! Press any key...\n");
 getch();

#endif

 return 0;
}
