/*
        @@@@@    @@@@    @@@@   @@  @@  @@   @@  @@@@   @@  @@
        @@  @@  @@  @@  @@  @@  @@ @@   @@@ @@@ @@  @@  @@@ @@
        @@@@@   @@@@@@  @@      @@@@    @@ @ @@ @@@@@@  @@ @@@
        @@      @@  @@  @@  @@  @@ @@   @@   @@ @@  @@  @@  @@
        @@      @@  @@   @@@@   @@  @@  @@   @@ @@  @@  @@  @@

        @@       @@@@    @@@@    @@@@   @@      @@@@@    @@@@    @@@@
        @@      @@  @@  @@      @@      @@      @@      @@      @@
        @@      @@  @@   @@@@    @@@@   @@      @@@@     @@@@    @@@@
        @@      @@  @@      @@      @@  @@      @@          @@      @@
        @@@@@@   @@@@    @@@@    @@@@   @@@@@@  @@@@@    @@@@    @@@@

        @@  @@  @@@@@@  @@@@@   @@@@@    @@@@
        @@  @@    @@    @@  @@  @@      @@  @@
        @@  @@    @@    @@  @@  @@@@    @@  @@
         @@@@     @@    @@  @@  @@      @@  @@
          @@    @@@@@@  @@@@@   @@@@@    @@@@

         @@@@    @@@@   @@@@@   @@@@@    @@@@
        @@  @@  @@  @@  @@  @@  @@      @@  @@
        @@      @@  @@  @@  @@  @@@@    @@
        @@  @@  @@  @@  @@  @@  @@      @@  @@
         @@@@    @@@@   @@@@@   @@@@@    @@@@

        PackMan Lossless Video EnCoder

        RGB => YCbCr => Frame Difference => WaveLet Transform => HUFFMAN -> RLE
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>

#define u8 unsigned char
#define s8 signed char

#define u16 unsigned short int
#define s16 signed short int

#define u32 unsigned long int
#define s32 signed long int

#include "huffman.h"
#include "rle.h"

#define W 160 /*   */
#define H 128 /*   */

#define NW 3  /*  WaveLet  */

#define KF 12 /*    */

#define COMPRESS 16384 /*     */

u8 nkf=0; //No Key Frame

u8 RGB_Buffer[H][W][3]; //  

s8 Frame[3][H][W];      // 
s8 Frame_Pred[3][H][W]; // 
s8 Frame_Temp[3][H][W]; // 

s8 CacheIN[W];  //  WaveLet 
s8 CacheOUT[W];

u8 *HUF_Buf;     //   Huffman
u8 RLE_Buf[H*W]; //   RLE

void YCbCrEncode(void)
{
 register u8 r,g,b;
 register s8 y,cb,cr;

 for(u32 j=0;j<H;j++)for(u32 i=0;i<W;i++)
 {
  r=RGB_Buffer[j][i][0]>>2; //Swap R<=>B & Convert 8:8:8 to 6:6:6 (for UG-6028 OLED Display)
  g=RGB_Buffer[j][i][1]>>2;
  b=RGB_Buffer[j][i][2]>>2;

  y =(s8)((r+(g<<1)+b)>>2); //Reversible Forward Color Transform
  cb=(s8)(b-g);
  cr=(s8)(r-g);

  Frame[2][j][i]=y;
  Frame[1][j][i]=cb;
  Frame[0][j][i]=cr;
 }
}

void LE_GALL_53_F(u32 n) //Le Gall 5/3 Forward WaveLet Transform(FWT)
{
 register u32 i;

 //Predict
 for(i=1;i<n-2;i+=2)CacheIN[i]-=(CacheIN[i-1]+CacheIN[i+1])/2;
 CacheIN[n-1]-=CacheIN[n-2];

 //Update
 for(i=2;i<n;i+=2)CacheIN[i]+=(CacheIN[i-1]+CacheIN[i+1])/4;
 CacheIN[0]+=CacheIN[1]/2;

 //Pack
 for(i=0;i<(n>>1);i++)
 {
  CacheOUT[i       ]=CacheIN[ i<<1   ];
  CacheOUT[i+(n>>1)]=CacheIN[(i<<1)|1];
 }
}

void FWT_X(u32 n)
{
 u32 x,y,c;
 for(c=0;c<3;c++)for(y=0;y<H>>n;y++)
 {
  for(x=0;x<W>>n;x++)CacheIN[x]=Frame[c][y][x];
  LE_GALL_53_F(W>>n);
  for(x=0;x<W>>n;x++)Frame[c][y][x]=CacheOUT[x];
 }
}

void FWT_Y(u32 n)
{
 u32 x,y,c;
 for(c=0;c<3;c++)for(x=0;x<W>>n;x++)
 {
  for(y=0;y<H>>n;y++)CacheIN[y]=Frame[c][y][x];
  LE_GALL_53_F(H>>n);
  for(y=0;y<H>>n;y++)Frame[c][y][x]=CacheOUT[y];
 }
}

int main(int argc,char** argv)
{
 u32 frame=0; // 
 u16 head;

 if(argc!=3)
 {
  printf("\nUsage: encoder InFile OutFile\n");
  return -1;
 }

 FILE *fin=fopen(argv[1],"rb");
 if(fin==NULL)return -1;

 FILE *fout=fopen(argv[2],"wb");
 if(fout==NULL)return -1;

 printf("\n");

 memset(Frame_Pred,0,sizeof(Frame_Pred));

 fputs("PackMan",fout); //PackMan\0 -   (8 )
 fputc(0,fout);

 Encode:

 if(kbhit())goto End; //   -  

 fread(RGB_Buffer,3,W*H,fin); // 

 if(feof(fin))goto End; //   -  

 YCbCrEncode(); //RGB => YCbCr

 memcpy(Frame_Temp,Frame,sizeof(Frame)); //Save Current Frame to Temporary Frame

 if(nkf)for(s32 c=2;c>=0;c--)for(u32 y=0;y<H;y++)for(u32 x=0;x<W;x++)Frame[c][y][x]-=Frame_Pred[c][y][x]; //KeyFrame ? Frame : Frame Difference

 memcpy(Frame_Pred,Frame_Temp,sizeof(Frame)); //Restore Temporary Frame to Previous Frame

 for(u32 n=0;n<NW;n++) //2D Forward WaveLet Transform
 {
  FWT_X(n);
  FWT_Y(n);
 }

 for(s32 c=2;c>=0;c--) //Y Cb Cr
 {
  u32 HUF_Len=CompressHuffman((u8*)&Frame[c][0][0],W*H,&HUF_Buf); //Huffman Coding
  u32 RLE_Len=CompressRLE(HUF_Buf,HUF_Len,RLE_Buf);               //RLE Coding

  if(RLE_Len<HUF_Len) // RLE   ,    RLE-
  {

   if(RLE_Len>COMPRESS) //    RLE
   {
    printf("Exceed Frame RLE %u > %u MAX! Press any key...\n",RLE_Len,COMPRESS);
    getch();
    goto End;
   }

   head=RLE_Len-1;         //bits(0..13)=Size-1
   head|=0x8000;           //bit (15   )=1 RLE(HUFFMAN)
   if(!nkf)head|=0x4000;   //bit (14   )=1 KeyFrame

   fwrite(&head,2,1,fout);         // Frame Header

   fwrite(RLE_Buf,1,RLE_Len,fout); //  

  }
  else //   HUF-
  {

   if(HUF_Len>COMPRESS) //    HUF
   {
    printf("Exceed Frame HUF %u > %u MAX! Press any key...\n",HUF_Len,COMPRESS);
    getch();
    goto End;
   }

   head=HUF_Len-1;         //bits(0..13)=Size-1
                           //bit (15   )=0 HUFFMAN
   if(!nkf)head|=0x4000;   //bit (14   )=1 KeyFrame

   fwrite(&head,2,1,fout);         // Frame Header

   fwrite(HUF_Buf,1,HUF_Len,fout); //  

  }

  free(HUF_Buf); // 
 }

 nkf++;
 if(nkf==KF)nkf=0; //  

 printf("Frame: %u \n",frame); //   
 frame++;

 goto Encode; //   

 End:

 //  -    ( )
 head=0;
 fwrite(&head,2,1,fout);

 fclose(fout);
 fclose(fin);

 printf("\nEncoding finished! Press any key...\n");
 getch();

 return 0;
}
